/*
 * This file is part of Alida, a Java library for 
 * Automatic Logging of Process Information in Data Analysis.
 *
 * Copyright (C) 2010 - 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 3751 $
 * $Date: 2011-05-03 10:13:56 +0200 (Di, 03 Mai 2011) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.Alida.operator;

import java.util.*;
import java.io.*;

import org.apache.xmlbeans.*;

/** Each instance of this class represents an operator invocation for the
 * implicit processing graph.
 * It holds input and output ports used to link the objects according to the data flow,
 * as well as the parameter's values upon invocation in a parameter hash.
 */
public class ALDOpNode {

	// Experimental
	/** if true all instances of ALDOpNode will be collected in a
	 * static Vector.
	 * WARNING: no instance of ALDOpNode will ever be freed currently in this case
	 */
	private static boolean collectInstances = false;

	// Maybe we should use weak references to allow instances of opNode to be freed
	private static Vector<ALDOpNode> allInstances = new Vector<ALDOpNode>();  

    static synchronized Vector<ALDOpNode> getAllInstances()  {  
         return( (Vector<ALDOpNode>)allInstances.clone() );  
    }  

	public static synchronized void printInstanceStatistics() {
		printInstanceStatistics( System.out);
	}

	public static synchronized void printInstanceStatistics( PrintStream stream) {
		stream.println( "Currently there are " + allInstances.size() +
						" instances of ALDOpNode");

		int numberParameters = 0;
		int charsInParameters = 0;
		int charsInValues = 0;

		for ( int i = 0 ; i < allInstances.size() ; i++ ) {
			Enumeration<String> keys = allInstances.elementAt( i).getParameterKeys();

			while(keys.hasMoreElements()) {
				numberParameters++;
				String key = keys.nextElement();
				charsInParameters += key.length();

				charsInValues += allInstances.elementAt( i).getParameter( key).length();
			}
		}
		stream.println( "   with a total of " + numberParameters + " parameters with " +
						+ charsInParameters + " characters in the keys and " +
						+ charsInValues + " characters in the values");
	}

	// END Experimental

	/** the input ports of this opNode.
	 */
	private ALDInputPort[] inputPorts; 

	/** the output ports of this opNode.
	 */
	private ALDOutputPort[] outputPorts; 

	/** name of the operator for which an invocations is represented by this opNode
	 */
	final private String name;			
	
	/** version of the operator for which an invocations is represented by this opNode
	 */
	final private String version;

	/** class of the operator for which an invocations is represented by this opNode
	 */
	final private Class operatorClass;

	/** This opnode is hidden, i.e. not include, in the processing history.
	 */
	boolean hidden; 

	/** This opnode prefers a complete DAG to constructed for its invocation.
	 */
	final boolean completeDAG;

	/** This hash contains the values of all parameters as return by the toString()
      * method of the parameters at the time of invocation of the operator sing runOp().
	  */

	private Hashtable<String, String> parameterHash;

	/** parameteHash as an XmlObject
	 */
	private XmlObject parameterHashAsXml; 

	/** parent of this opNode, will be set during back tracing
	 */
	private ALDOpNode parent; 

	/** children of this opNode, will be set during back tracing 
	 */
	private Vector<ALDOpNode> children; 

	/** children of this opNode, set when calling operators by runOp
	 */
	private Vector<ALDOpNode> directlyRegisteredChildren; 

	/** depth in calling stack of Operators, set when back tracing 
	 */
	private int depth; 

	/** ALDDataPorts created within this opNode, set during back tracing 
	 */
	private Vector<ALDDataPort> includedData; 

	/** Construct a opnode for the operator op and hidden flag
	 */
	public ALDOpNode( ALDOperator op, boolean hidden) {
		//Experimental
		if ( collectInstances ) {
        	allInstances.add( this );  
		}
		//END Experimental

		this.completeDAG = op.completeDAG;
		this.depth = Integer.MAX_VALUE;
		this.hidden = hidden;

		this.name = op.getName();
		this.operatorClass = op.getClass();
		this.version = op.getVersion();

		try { 
			inputPorts = new ALDInputPort[ op.getNumInputs()];
			String[] inputNames = op.getInputNames();
			for ( int i=0 ; i < op.getNumInputs() ; i++ )
				inputPorts[i] = new ALDInputPort( this, i, inputNames[i]);
	
			outputPorts = new ALDOutputPort[ op.getNumOutputs()];
			String[] outputNames = op.getOutputNames();
			for ( int i=0 ; i < op.getNumOutputs() ; i++ )
				outputPorts[i] = new ALDOutputPort( this, i, outputNames[i]);
	
			parameterHash = new Hashtable<String, String>();
	
			children = new Vector<ALDOpNode>();
			directlyRegisteredChildren = new Vector<ALDOpNode>();
			includedData = new Vector<ALDDataPort>();

			// copy the parameters of the descriptor to a private hash
			// this is necessary, as the descriptor may be reused lateron with different
			// values of its parameters (or the values changes for another purpose)
			String[] parameterNames = op.getParameterNames();
			for ( int n = 0 ; n < op.getNumParameters() ; n++ ) {
				if ( op.getParameter( parameterNames[n]) != null ) 
					parameterHash.put( op.getParameterDescriptor( parameterNames[n]).name, 
								op.getParameter( parameterNames[n]).toString());
			}
			//parameterHashAsXml = op.parametersToXmlObject();
		} catch (Exception e) {
			e.printStackTrace();
		} 
	}

	/** Return the name of the operator represented by this opNode
	 */
	public String getName() {
		return name;
	}

	/** Get the software version
	 */
	public String getVersion() {
		return version;
	}

	/** Return the class object of the operator represented by this opNode
	 */
	public Class getOperatorClass() {
		return operatorClass;
	}

	/** Set depth 
	 */
	void setDepth( int depth) {
		this.depth = depth;
	}

	/** Get depth 
	 */
	public int getDepth() {
		return depth;
	}

	/** Add a child found during back tracing
	 */
	void addChild( ALDOpNode opNode) {
		children.add( opNode);
	}

	/** Add a child during directly during invocation
	 */
	void addDirectChild( ALDOpNode opNode) {
		directlyRegisteredChildren.add( opNode);
	}

	/** Add a data port found during back tracing
	 */
	void addData( ALDDataPort data) {
		includedData.add( data);
	}

	/** Get input port with index i of this opNode
	 */
	public ALDPort getInputPort( int i) {
		return inputPorts[i];
	}

	/** Get output port with index i of this opNode
	 */
	public ALDOutputPort getOutputPort( int i) {
		return outputPorts[i];
	}

	/** Set the origin of the input port with index i.
      * This sets the origin of the associated input port as well
      * as its canoncial classname. The explanation is usually to
      * be copied from the argument descriptor of this input port.
      *
      * @param	i index of the input port
      * @param	obj object to assiciate with the port
      * @param	explanation explnatory string of port
     **/
	void setInOrigin( int i, Object obj, String explanation) {
		if ( obj != null ) {
			inputPorts[i].setOrigin(ALDOperator.getALDPortHashAccessKey().getHistoryLink(obj));
			if ( obj.getClass().getPackage() == null )
				inputPorts[i].setClassname( obj.getClass().getName());
			else
				inputPorts[i].setClassname( obj.getClass().getPackage()+"."+obj.getClass().getName());
		}
		inputPorts[i].setExplanation( explanation);

	}

	/** Set the origin of the output port with index i.
      * This sets the origin of the accociated output port as well
      * as its canoncial classname. The explanation is usually to
      * be copied from the argument descriptor of this output port.
      *
      * @param	i index of the input port
      * @param	obj object to assiciate with the port
      * @param	explanation explnatory string of port
     **/
	void setOutOrigin( int i, Object obj, String explanation) {
		if ( obj != null ) {
			outputPorts[i].setOrigin( ALDOperator.getALDPortHashAccessKey().getHistoryLink(obj) );
			if ( obj.getClass().getPackage() == null ) {
				outputPorts[i].setClassname( obj.getClass().getName());
			} else {
				outputPorts[i].setClassname( obj.getClass().getPackage()+"."+obj.getClass().getName());
			}
		}
		outputPorts[i].setExplanation( explanation);
	}

	/** Set the parent opNode
	 */
	public void setParent( ALDOpNode parent) {
		this.parent = parent;
	}

	/** Get the parent opNode
	 */
	public ALDOpNode getParent() {
		return parent;
	}

	/** Get the all opNode children found during back tracing
	 */
	protected Vector<ALDOpNode> getChildren() {
		return children;
	}

	/** Get the all direct opNode children registered during invocation
	 */
	protected Vector<ALDOpNode> getDirectlyRegisteredChildern() {
		return directlyRegisteredChildren;
	}

	/** Return all included data ports found during back tracing
     */
	protected Vector<ALDDataPort> getIncludedData() {
		return includedData;
	}

	/** Return all input ports of this opNode
     */
	protected ALDInputPort[] getInputPorts() {
		return inputPorts;
	}

	/** Return all output ports of this opNode
     */
	protected ALDOutputPort[] getOutputPorts() {
		return outputPorts;
	}

	/** Return all keys of the parameter hash
     */
	public Enumeration<String> getParameterKeys() {
		return this.parameterHash.keys();
	}

	/** Return value of parameter for given key
     */
	public String getParameter( String key) {
		return this.parameterHash.get( key);
	}

	/** Set hidden flag of this opNode.
	 * This prevents this opNode to be included into a processing
	 * history as explicitly constructed.
     */
	public void setHidden( boolean hidden) {
		this.hidden = hidden;
	}

	/** Get hidden flag of this opNode.
	 * If set this opNode is not included into a
	 * history as explicitly constructed.
     */
	public boolean getHidden() {
		return this.hidden;
	}

	/** Returns the parameter hash where values of parameters
	 * are represented as xml objects.
	 */
	public XmlObject getParameterHashAsXml() {
		return this.parameterHashAsXml;
	}

	/** Print information if this opNode to standard output.
	 */
	public void print() {
		System.out.println( "ALDOpNode ( with " + inputPorts.length +
					"/" + outputPorts.length + " ports for operator " + name +
					" at depth " + depth + ", completeDAG " + completeDAG + ")" );
		System.out.println( "    >>>>>   children: ");
		Iterator<ALDOpNode> oItr = children.iterator();
		while ( oItr.hasNext() ) {
			System.out.println( "    " + oItr.next().getName());
		}
		System.out.println( "    <<<<<   children: ");
		System.out.println( "    >>>>>   direct children: ");
		oItr = directlyRegisteredChildren.iterator();
		while ( oItr.hasNext() ) {
			System.out.println( "    " + oItr.next().getName());
		}
		System.out.println( "    <<<<<   direct children: ");
	}

}
