/*
 * This file is part of Alida, a Java library for 
 * Automatic Logging of Process Information in Data Analysis.
 *
 * Copyright (C) 2010 - 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 3758 $
 * $Date: 2011-05-03 17:17:23 +0200 (Di, 03 Mai 2011) $
 * $Author: posch $
 * 
 */
//TODO: do we need unset..... if we  wompletely switch to annotations??

package de.unihalle.informatik.Alida.operator;

import de.unihalle.informatik.Alida.version.ALDVersionProvider;
import de.unihalle.informatik.Alida.version.ALDVersionProviderFactory;
import de.unihalle.informatik.Alida.exceptions.*;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.OperatorExceptionType;
import de.unihalle.informatik.Alida.operator.ALDArgumentAnnotation.ALDArgumentType;

import com.thoughtworks.xstream.*;
import com.thoughtworks.xstream.io.xml.DomDriver;

import org.apache.xmlbeans.*;
import org.graphdrawing.graphml.xmlns.GraphmlDocument;

import java.io.*;
import java.util.*;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.annotation.Annotation;

/** This is the abstract super class for all Alida operators, 
  * automatically handling the logging of the processing history.
  *<p>
  * The interface consisting of parameters, inputs, outputs, and supplemental arguments
  * needs to be define by annotating the corresponding fields (i.e. member variables) 
  * using the <code>@Parameter</code>
  * annotation.
  * Only the attributes <code>required</code>, <code>output</code>, and <code>description</code> of this annotations are used for operators.
  * For input and output primitive data types, interned strings, and 
  * cached numerical objects are not allowed .
  * This is required to ensure the history mechanism to work properly.
  * Values of parameters and inputs have to be set prior to invoking the operator via runOp()
  * and the resulting outputs may be retrieved
  * from the operator after return from runOp().
  * In addition to parameters, inputs, and outputs an operator may use
  * supplemental arguments e.g. to control output or debugging information
  * as well as to return intermediate results. The outputs of the operator are expected to
  * be independent of the value of these supplemental values which are not stored in the processing history.
  * <p>
  * A defaul constructor without arguments has to be implemented for all
  * features of code generation to be available.
  * <p>
  * The method operate() implements the processing of this operator. All information
  * passed into and back from the operator are passed via the member variables of the operator
  * which are properly annotated.
  * This method should by no means be used to invoke processing of the operator directly,
  * rather  runOp() is to be used.
  * <p> 
  * The method runOp() is called by the user of an operator to invoke processing.
  * runOp first checks if all required parameters and inputs are set to non-null values
  * and subsequently check the validity of parameters and inputs
  * as defined by the method <code>validateCustom</code>.
  * <p>
  * Refer to the Alida manual for more details and examples.
  * 
  * @author posch
 */

abstract public class ALDOperator {
	/** Verbose flag to be inherited by all operators.
	 */
	@Parameter( label= "verbose", type = Parameter.Type.SUPPLEMENTAL, description = "Verbose flag")
	protected Boolean verbose = new Boolean(false);

	// ====================================================
	// This is currently experimental for debugging purposes
	/** Mode of implicit construction of the processing graph.
	 * Still experimental at this point!!
	 * 0 = regular construction
	 * 1 = do not create opNodes for hidden operator invocations
	 * 2 = prevent implicit construction completely
	 */
	private int constructionMode = 0;


	// END of This is currently experimental for debugging purposes
	// ====================================================

	/**
	 * Get access to the central port hash.
	 */
	protected static ALDPortHashAccess portHashAccess = new ALDPortHashAccess();
	
	/**
	 * Get an instance of a class providing version information.
	 */
	protected ALDVersionProvider versionProvider = 
		ALDVersionProviderFactory.getProviderInstance();
	
	/**
	 * Returns a reference to the port hash access object.
	 */
	public static ALDPortHashAccess getALDPortHashAccessKey() {
		return portHashAccess;
	}
	
	/**
	 * Reads the history graph of the given object from file.
	 * 
	 * @param obj			Object for which the history is to be read.
	 * @param filename	File from where the history is to be read.
	 */
	public static void readHistory(Object obj, String filename) {
		portHashAccess.readHistory(obj, filename);
	}
	
	/** 
	 * Write the processing history if any to a graphml file.
	 * Equivalent to <code>writeHistory(obj, filename, ALDProcessingDAG.HistoryType.COMPLETE, false)</code>
	 */
	public static void writeHistory(Object obj, String filename) 
		throws ALDProcessingDAGException,ALDOperatorException {
		writeHistory(obj, filename, ALDProcessingDAG.HistoryType.COMPLETE, false);
	}

	/** 
	 * Write the processing history if any to a graphml file.
	 * Equivalent to <code>writeHistory(obj, filename, historyType, false)</code>
	 */
	public static void writeHistory(Object obj, String filename, ALDProcessingDAG.HistoryType historyType) 
	throws ALDProcessingDAGException,ALDOperatorException {

		writeHistory(obj, filename, historyType, false);
	}

	/** 
	 * Write the processing history if any to a graphml file.
	 *
	 * @param	obj	Object for which the history to write for.
	 * @param	filename	Filename to write the processing history into. 
	 * 									(as graphml/XML file)
	 * @param	historyType	Type/mode of the history.
	 * @param	ignoreHiding	If true, hiding of opNodes is ignored.
	 */
	public static void writeHistory(Object obj, String filename,
			ALDProcessingDAG.HistoryType historyType,	boolean ignoreHiding) 
	throws ALDProcessingDAGException,ALDOperatorException {
		portHashAccess.writeHistory(obj, filename, historyType, ignoreHiding);
	}

	/** Does this operator prefer a complete DAG or a data dependency DAG.
	 * This field may be changed when implementing an operator.
	 */
	protected boolean completeDAG = true;

	// really only debugging
	private boolean debug = false;

	/** A weak hashmap which contains for each thread a stack of opNodes reflecting
	 * the current state of the method stack.
	 *  Whenever runOp() is called, an opNode is created which is associated with
	 *  this call. It is pushed onto the stack and popped at the end of runOp().
	 * As the first element a opNode for the dummy operator ALDToplevelOperator
	 * is created which represents outer or top most level of operator calls.
	 */
	static WeakHashMap<Thread,Stack<ALDOpNode>> opNodeStackHash= new WeakHashMap<Thread,Stack<ALDOpNode>>();

	/** Name of the operator
	 */
	public final String name;

	/** svn version control
	 */
//	public final String version; 

	// descriptors collected from this operator and all its super classes
	private Hashtable<String,ALDOpArgumentDescriptor> parameterDescriptorsAll;
	private Hashtable<String,ALDOpArgumentDescriptor> inputDescriptorsAll;
	private Hashtable<String,ALDOpArgumentDescriptor> outputDescriptorsAll;
	private Hashtable<String,ALDOpArgumentDescriptor> supplementalDescriptorsAll;

	/** This constructor initializes an operator.
	 *  Name and version are retrieved and hashes are initialized.
	 */
	protected ALDOperator()  throws ALDOperatorException {
		this.name = this.getClass().getSimpleName();

		this.parameterDescriptorsAll = new Hashtable<String,ALDOpArgumentDescriptor>();
		this.inputDescriptorsAll = new Hashtable<String,ALDOpArgumentDescriptor>();
		this.outputDescriptorsAll = new Hashtable<String,ALDOpArgumentDescriptor>();
		this.supplementalDescriptorsAll = new Hashtable<String,ALDOpArgumentDescriptor>();

		// loop for this class and all super classes over all declared fields to find Annotations
		Class<?> myclass = this.getClass();
		Class currentClass = myclass;
		do {
        
			for ( Field field : myclass.getDeclaredFields() ) {
				String name = field.getName();

				Parameter pAnnotation = field.getAnnotation( Parameter.class);

				// if field is annotated as parameter create and add appropriate descriptor
				// and not already decleared in a class downstream of the inheritance hierachy
				// (as we add the first field for each name we find when walking up the inheritance hierachy
				if ( pAnnotation != null && !fieldContained( name) ) {
					ALDArgumentAnnotation mAnnotation = field.getAnnotation( ALDArgumentAnnotation.class);
					Object value = null;

					String explanation;
					if ( mAnnotation == null )
						explanation = pAnnotation.description();
					else
						explanation = mAnnotation.description();

					ALDOpArgumentDescriptor argumentDescriptor =
						new ALDOpArgumentDescriptor( name, field.getType(), 
													explanation,
													pAnnotation.required(), 
													value, field);

					Parameter.Type	myType;
					if ( mAnnotation == null ) {
						myType = pAnnotation.type();
					} else {
						if ( mAnnotation.type() == ALDArgumentAnnotation.ALDArgumentType.INPUT )
							myType = Parameter.Type.INPUT;
						else if ( mAnnotation.type() == ALDArgumentAnnotation.ALDArgumentType.OUTPUT )
							myType = Parameter.Type.OUTPUT;
						else if ( mAnnotation.type() == ALDArgumentAnnotation.ALDArgumentType.PARAMETER )
							myType = Parameter.Type.PARAMETER;
						else if ( mAnnotation.type() == ALDArgumentAnnotation.ALDArgumentType.SUPPLEMENTAL )
							myType = Parameter.Type.SUPPLEMENTAL;
						else {
							myType = null;
							throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNSPECIFIED_ERROR, 
										"<" + field.getName() + "> of class <" + field.getType().getName() + ">"+
                                    " in operator " + this.name + " annotates illeagal type " +
								mAnnotation.type() );
						}
					}

					if ( myType == Parameter.Type.OUTPUT ) {
						// output
						if ( ! allowedClassForIO( field.getType()) ) {
							throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.INVALID_CLASS, 
										"<" + field.getName() + "> of class <" + field.getType().getName() + ">"+
                                    " in operator " + this.name);
						}
						if ( this.debug ) {
							System.out.println( "   ALDOperator():: adding output argument <" + name + ">");
						}
						this.outputDescriptorsAll.put( name, argumentDescriptor);


					} else if ( myType == Parameter.Type.INPUT ) {
						// input
						if ( ! allowedClassForIO( field.getType()) ) {
							throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.INVALID_CLASS, 
									"<" + field.getName() + "> of class <" + field.getType().getName() + ">" +
									" in operator " + this.name);
						}
						if ( this.debug ) {
							System.out.println( "   ALDOperator():: adding input argument <" + name + ">" +
								" of type <" + field.getType() + ">");
						}
						this.inputDescriptorsAll.put( name, argumentDescriptor);

					} else if ( myType == Parameter.Type.PARAMETER ) {
						// parameter
						if ( this.debug ) {
							System.out.println( "   ALDOperator():: adding parameter argument <" + name + ">");
						}
						this.parameterDescriptorsAll.put( name, argumentDescriptor);
					} else if ( myType == Parameter.Type.SUPPLEMENTAL ) {
						// supplemental
						if ( this.debug ) {
							System.out.println( "   ALDOperator():: adding supplemental argument <" + name + ">");
						}
						this.supplementalDescriptorsAll.put( name, argumentDescriptor);
					}
				}
			} 

			myclass = myclass.getSuperclass();
        } while ( myclass != null );
	}

	/** Get the name of this operator 
	 * @return name of the operator
	 */
	public final String getName() {
		return this.name;
	}

	/** Get the version in SVN of this operator 
	 * @return version of the operator
	 */
	public final String getVersion() {
//		return this.version;
		if (this.versionProvider != null) {
			return this.versionProvider.getVersion();
		}
		return "unknown";
	}

	// -------------------------------------------------------
	// access functions to parameters

	/** Get the number of parameters
	 * @return number of parameters
	 */
	public final int getNumParameters() {
		return this.parameterDescriptorsAll.size();
	}

	/** Get the names of parameters
	 * @return array of parameter names
	 */
	public final String[] getParameterNames() {
		String[] keyArray = new String[getNumParameters()];
		Enumeration<String> keys = this.parameterDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the parameter descriptor for given name.
	 * @param	name	Name of the parameter to get the new value for
	 * @return	descriptor value 
	 */
	@Deprecated
	public final ALDOpArgumentDescriptor getParameterDescriptor( String name) 
		throws ALDOperatorException {
		if ( ! this.parameterDescriptorsAll.containsKey( name) )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return this.parameterDescriptorsAll.get( name);
	}

	/** Get the value of a parameter specified by name. 
	 * If a parameter is of type Boolean and is unset, than False is returned,
     * for all other types null is return for unset parameters.
	 *
	 * @param	name	Name of the parameter to get the new value for
	 * @return	If the parameter is set return its value. If the parameter is unset return False for parameter of type Boolean,
	 *				null otherwise.
	 */
	final Object getParameter(String name) 
		throws ALDOperatorException {
		if ( ! this.parameterDescriptorsAll.containsKey( name)  )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);

		try {
			Field field = getParameterDescriptor( name).field;
			field.setAccessible(true);
			return field.get( this);
		} catch ( IllegalAccessException e ) {
            e.printStackTrace();
        }
		return null;
	}

	/** Set the value of a parameter specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of a sub class of the specified parameter or
     * the standard widening for primitive data types apply to the numeric wrapper class.
	 *
	 * @param	name	Name of the parameter to set a new value for
	 * @param	value	new value
	 */
	final void setParameter(String name, Object value) 
		throws ALDOperatorException {
		if ( value != null ) {
			if ( ! this.parameterDescriptorsAll.containsKey( name)  )
				throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);

			try {
				Field field = getParameterDescriptor( name).field;
				field.setAccessible(true);
				field.set( this, value);
			} catch ( IllegalAccessException e ) {
            	e.printStackTrace();
        	}
		}
	}
	
	// -------------------------------------------------------
	// access functions to inputs

	/** Get the number of inputs
	 * @return number of inputs
	 */
	public final int getNumInputs() {
		return this.inputDescriptorsAll.size();
	}

	/** Get the names of inputs
	 * @return array of input names
	 */
	public final String[] getInputNames() {
		String[] keyArray = new String[getNumInputs()];
		Enumeration<String> keys = this.inputDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the input descriptor for given name.
	 * @param	name	Name of the input to get the new value for
	 * @return	descriptor value 
	 */
	@Deprecated
	public final ALDOpArgumentDescriptor getInputDescriptor( String name) 
		throws ALDOperatorException {
		if ( ! this.inputDescriptorsAll.containsKey( name) )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return this.inputDescriptorsAll.get( name);
	}

	/** Get the value of an input specified by name. 
	 *
	 * @param	name	Name of the input to get the new value for
	 * @return	input value or null unset
	 */
	final Object getInput(String name) 
		throws ALDOperatorException {
		if ( ! this.inputDescriptorsAll.containsKey( name)  )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);

		try {
			Field field = getInputDescriptor( name).field;
			field.setAccessible(true);
			return field.get( this);
		} catch ( IllegalAccessException e ) {
            e.printStackTrace();
        }
		return null;
	}

	/** Set the value of an input specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of the input specified.
	 *
	 * @param	name	Name of the input to set a new value for
	 * @param	value	new value
	 */
	final void setInput(String name, Object value) 
		throws ALDOperatorException {
		if ( value != null ) {
			if ( ! this.inputDescriptorsAll.containsKey( name)  )
				throw new ALDOperatorException(
						ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			try {
				Field field = getInputDescriptor( name).field;
				field.setAccessible(true);
				field.set( this, value);
			} catch ( IllegalAccessException e ) {
            	e.printStackTrace();
			}
		}
	}
		
	// -------------------------------------------------------
	// access functions to outputs

	/** Get the number of outputs
	 * @return number of outputparameters
	 */
	public final int getNumOutputs() {
		return this.outputDescriptorsAll.size();
	}

	/** Get the names of outputs
	 * @return array of output names
	 */
	public final String[] getOutputNames() {
		String[] keyArray = new String[getNumOutputs()];
		Enumeration<String> keys = this.outputDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the output descriptor for given name.
	 * @param	name	Name of the output to get the new value for
	 * @return	descriptor value 
	 */
	@Deprecated
	public final ALDOpArgumentDescriptor getOutputDescriptor( String name) 
		throws ALDOperatorException {
		if ( ! this.outputDescriptorsAll.containsKey( name) )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return this.outputDescriptorsAll.get( name);
	}

	/** Get the value of a output specified by name. 
	 *
	 * @param	name	Name of the output to get the new value for
	 * @return	output descriptor or null if unset
	 */
	final Object getOutput(String name) 
		throws ALDOperatorException {
		if ( ! this.outputDescriptorsAll.containsKey( name)  )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);

		try {
			Field field = getOutputDescriptor( name).field;
			field.setAccessible(true);
			return field.get( this);
		} catch ( IllegalAccessException e ) {
            e.printStackTrace();
        }
		return null;
	}

	/** Set the value of an output specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of the output specified.
	 *
	 * @param	name	Name of the output to set a new value for
	 * @param	value	new value
	 */
	final void setOutput(String name, Object value) 
		throws ALDOperatorException {
		if ( value != null ) {
			if ( ! this.outputDescriptorsAll.containsKey( name)  )
				throw new ALDOperatorException(
						ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			try {
				Field field = getOutputDescriptor( name).field;
				field.setAccessible(true);
				field.set( this, value);
			} catch ( IllegalAccessException e ) {
            	e.printStackTrace();
        	}
		}
	}
	
	// -------------------------------------------------------
	// access functions to supplemental values

	/** Get the number of supplementals
	 * @return number of supplementals
	 */
	public final int getNumSupplementals() {
		return this.supplementalDescriptorsAll.size();
	}

	/** Get the names of supplementals
	 * @return array of supplemental names
	 */
	public final String[] getSupplementalNames() {
		String[] keyArray = new String[getNumSupplementals()];
		Enumeration<String> keys = this.supplementalDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the supplemental descriptor for given name.
	 * @param	name	Name of the supplemental to get the new value for
	 * @return	descriptor value 
	 */
	@Deprecated
	public final ALDOpArgumentDescriptor getSupplementalDescriptor( String name) 
		throws ALDOperatorException {
		if ( ! this.supplementalDescriptorsAll.containsKey( name) )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return this.supplementalDescriptorsAll.get( name);
	}

	/** Get the value of a supplemental specified by name. 
	 *  If a parameter is of type Boolean and is unset, than False is returned,
	 *  for all other types null is return for unset parameters.
	 *
	 * @param	name	Name of the supplemental to get the new value for
	 * @return	If the supplemental is set return its value. If the supplemental is unset return False for supplemental of type Boolean,
	 *			null otherwise.
	 */
	final Object getSupplemental(String name) 
		throws ALDOperatorException {
		if ( ! this.supplementalDescriptorsAll.containsKey( name)  )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		try {
			Field field = getSupplementalDescriptor( name).field;
			field.setAccessible(true);
			return field.get( this);
		} catch ( IllegalAccessException e ) {
            e.printStackTrace();
        }
		return null;
	}

	/** Set the value of a supplemental specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of a sub class of the specified supplemental or
     * the standard widening for primitive data types apply to the numeric wrapper class.
	 *
	 * @param	name	Name of the supplemental to set a new value for
	 * @param	value	new value
	 */
	final void setSupplemental(String name, Object value) 
		throws ALDOperatorException {
		if ( value != null ) {
			if ( ! this.supplementalDescriptorsAll.containsKey( name)  )
				throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			try {
				Field field = getSupplementalDescriptor( name).field;
				field.setAccessible(true);
				field.set( this, value);
			} catch ( IllegalAccessException e ) {
            	e.printStackTrace();
			}
		}
	}

	//==========================

	/** A legal method to invoke the operator and handles everything 
	 * necessary to protocol the processing history.
	 * This call has the same effect as <code>runOp( false)</code>, i.e.
	 * runs the operator as not hidden.
	 */
	public final void runOp( ALDOperator callingOperator) throws ALDOperatorException,ALDProcessingDAGException {
		runOp( false);
	}

	/** A legal method to invoke the operator and handles everything 
     * necessary to protocol the processing history.
     * This call has the same effect as <code>runOp( false)</code>, i.e.
     * runs the operator as not hidden.
     */
	@Deprecated
	public final void runOp() throws ALDOperatorException,ALDProcessingDAGException {
		runOp( false);
	}

	/** A legal method to invoke the operator and handles everything 
	 * necessary to protocol the processing history.
	 * Prior to calling <code>runOp</code>, input data and parameters have to be set.
	 * When called <code>runOp</code> validates the parameters and inputs.
	 * Validity requires for all operators, that all required parameters and inputs
	 * have non null values.
	 * In addition the implementation of an operator may impose further constrains
	 * defined by the method <code>validateCustom</code>
	 * which e.g. may restrict the interval of numerical parameters.
	 * After successful validation the <code>operate</code> method is called to
	 * do the actuall processing of this operator.
	 * Upon return from <code>runOp</code>, resulting output and supplemental data
	 * may be retrieved from the operator object.
	 * <br>
	 * If <code>hidden</code> is true, the activation of this operator will
	 * not be visible in the processing history.
	 *
	 * @param   hidden    The activation of this operator is hidden in the processing history..
	 */


	public final void runOp( boolean hidden) throws ALDOperatorException,ALDProcessingDAGException {

        if ( getVerbose().booleanValue() || this.debug ) {
            System.out.println( "");
            System.out.println( "Start of runOp");
            this.print();
        }
 
		validate();

		// find parent opNode via opNode stack of this thread
		Stack<ALDOpNode>  currentThreadsOpNodeStack = findCurrentThreadsOpNodeStack();
		ALDOpNode parentViaThread = currentThreadsOpNodeStack.peek();

		ALDOpNode opNode = null;
		ALDPort origPorts[] = null;
		String[] inputNames = null;

		if ( constructionMode == 0  || 
			 ( constructionMode == 1 && 
               ! hidden &&
			   ! parentViaThread.getHidden() ) )  {

			// create an opNode for this operation in the DAG
			opNode = new ALDOpNode( this, hidden);


			// find parent opNode via opNode stack of calling operator
			// dprecated
			// ALDOpNode parentViaOperator = findParentOpnodeViaOperator( callingOperator, opNode);
			//
			// register via operator
        	// if ( parentViaOperator != null ) 
			//		parentViaOperator.addDirectChild( opNode);

			// register this opNode to its parent opNode via thread parent
			parentViaThread.addDirectChild( opNode);
			opNode.setParent( parentViaThread);
			if ( this.debug || getVerbose() ) {
				System.out.println( "ALDOperator::runOp adding direct registered child " + this + "(" + opNode + ")\n" +
						"                   to opNode " + parentViaThread + "(" + parentViaThread.getName() + ")");
			}

			if ( parentViaThread.getHidden() ) 
				opNode.setHidden( true);

			currentThreadsOpNodeStack.push( opNode);

			// depreacted
			// push AFTER we added the child to handle recursive calls to an operator instance
			//this.currentOpNodeStack.push( opNode);


			// handle inputs
			origPorts = new ALDPort[this.getNumInputs()];
			inputNames = getInputNames();

			// first we store the original dataport of inputs
			for ( int i = 0; i < this.getNumInputs() ; i++ ) {
				Object input = this.getInput( inputNames[i]);
	
				if ( input != null ) {
	
					// register the input object
					if (!portHashAccess.isRegistered(input)) 
						portHashAccess.register(input);				
	
					// connect each input port of this opNode to the port
					// of the input data the operator was supplied with
					opNode.setInOrigin( i, input, getInputDescriptor( inputNames[i]).explanation);
	
					origPorts[i] = portHashAccess.getHistoryLink(input);
				} else {
					opNode.setInOrigin( i, null, getInputDescriptor( inputNames[i]).explanation);
					origPorts[i] = null;
				}
			}

			// and now we can set the dataports of the data to the input ports
			// (we have to do this in two for-loops to handle multiple inputs with
			// indentical ALDData objects correctly)
			for ( int i = 0; i < this.getNumInputs() ; i++ ) {
				Object input = this.getInput( inputNames[i]);

				if ( input != null ) {
					// set the output port of the input to the input port of this opNode
					portHashAccess.setHistoryLink(input, opNode.getInputPort( i));
				}
			}

			if ( getVerbose().booleanValue() || this.debug ) {
				System.out.println( "");
				System.out.println( "  before operate");
				for ( int i = 0; i < this.getNumInputs() ; i++ ) {
					System.out.println( "    port of " + i + "-th input " + opNode.getInputPort( i) +
							" with origin " + opNode.getInputPort( i).getOrigin() );
				}
			}
		}

		if ( getVerbose().booleanValue() || this.debug )
			System.out.println("Running operate...");

		// do the work
		operate();

		if (getVerbose().booleanValue() || this.debug )
			System.out.println("Running operate...done.");

		if ( constructionMode == 0  || 
			 ( constructionMode == 1 && 
               ! hidden &&
			   ! parentViaThread.getHidden() ) )  {

			// handle output ports: connect output ports of opNode to the passing output arguments
			String[] outputNames = getOutputNames();
			for ( int i = 0; i < this.getNumOutputs() ; i++ ) {
				Object output = this.getOutput( outputNames[i]);

				// check if datum has already a port
				if (output != null && !portHashAccess.isRegistered(output)) 
					portHashAccess.register(output);			

				if ( output != null ) {
				
					// register the output object to the history DB
					if (!portHashAccess.isRegistered(output)) 
						portHashAccess.register(output);

					// connect each output port of this opNode to the output port
					// its results (i.e. oututs) originate from
					opNode.setOutOrigin( i, output, getOutputDescriptor( outputNames[i]).explanation);

					// store properties of output at the output port
					if (output instanceof ALDData) {
						Enumeration<String> keys = ((ALDData)output).getPropertyKeys();
						while ( keys.hasMoreElements() ) {
							String key = keys.nextElement();
							opNode.getOutputPort( i).setProperty( key, ((ALDData)output).getProperty( key));
						}
					}
				} else {
					opNode.setOutOrigin( i, null, getOutputDescriptor( outputNames[i]).explanation);
				}
			}
			// restore port of inputs
			for ( int i = 0; i < this.getNumInputs() ; i++ ) {
				if ( origPorts[i] != null ) 
					portHashAccess.setHistoryLink(this.getInput(inputNames[i]),origPorts[i]);
			}

			// handle outputs: connected leaving outputs arguments to output ports of this opNode
			for ( int i = 0; i < this.getNumOutputs() ; i++ ) {
				Object output = this.getOutput( outputNames[i]);
	
				if ( output != null ) {
					// set the output port of the output (i.e. result) of the operator
					// to the output port of the opNode
					portHashAccess.setHistoryLink(output, opNode.getOutputPort(i));
				}
			}

			if ( getVerbose() != null && getVerbose().booleanValue() ) {
				System.out.println( "  End of runOp");
				this.print();
				for ( int i = 0; i < this.getNumOutputs() ; i++ ) {
					System.out.println( "    port for " + i + "-th result " + opNode.getOutputPort( i) +
						" with origin " + opNode.getOutputPort( i).getOrigin() );
					if ( opNode.getOutputPort( i).getOrigin() != null )
						opNode.getOutputPort( i).getOrigin().print();
				}
			}

			currentThreadsOpNodeStack.pop();
		}

		// deprecated
		// pop opNode from stack
		//this.currentOpNodeStack.pop();
	}

	// ===================================================================
	// Helper for runOp()

	/** find the opNode stack for the current thread.
	 * Create a new one, if not existing
	 */
	private Stack<ALDOpNode> findCurrentThreadsOpNodeStack() {
        Thread currentThread = Thread.currentThread();
        Stack<ALDOpNode> currentThreadsOpNodeStack = opNodeStackHash.get(currentThread);
        if ( currentThreadsOpNodeStack == null ) {
            try {
                if ( this.debug )
                    System.out.println( "   ALDOperator::runOp create opNodeStack for thread " +
                        currentThread);
                currentThreadsOpNodeStack = new Stack<ALDOpNode>();
                opNodeStackHash.put( currentThread, currentThreadsOpNodeStack);
                currentThreadsOpNodeStack.push( new ALDOpNode( new ALDToplevelOperator(), false));
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

		if ( this.debug ) 
			System.out.println( "   ALDOperator::runOp opNodeStack for thread " + currentThread + " found: " + currentThreadsOpNodeStack);
		return currentThreadsOpNodeStack;
    }

	// ===================================================================

	/** Validates the parameters and inputs of this operator.
	 *  The generic part of this validation checks that each required parameter
	 *  and input is set to a non-null value.
	 *  The custom part of validation is operator depended and implemented
	 * via the method <code>validateCustom</code>.
	 *  An ALDOperatorException is thrown if either validation fails.
 	 */
	public final void validate() throws ALDOperatorException {
		validateGeneric();
		validateCustom();
	}

	/** Generic validation of the parameters and inputs of this operator.
	 *  Validation succeeds if each required parameter
	 *  and input is set to a non-null value or supplies a default value.
	 *  An ALDOperatorException is thrown if validation fails.
 	 */
	public final void validateGeneric() throws ALDOperatorException {
		
		String[] inputNames = getInputNames();
		for ( int i = 0; i < this.getNumInputs() ; i++ ) {
			if ( this.inputDescriptorsAll.get( inputNames[i]).required &&
					getInput( inputNames[i]) == null &&
					this.inputDescriptorsAll.get( inputNames[i]).defaultValue  == null ) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED, 
												"ALDOperator.validateGeneric(): Required input argument\"" + inputNames[i] + "\""+
                                    " is null in operator " + this.name);  
			}            
		}

		String[] parameterNames = getParameterNames();
		for ( int i = 0; i < this.getNumParameters() ; i++ ) {
			if ( this.parameterDescriptorsAll.get( parameterNames[i]).required &&
					getParameter( parameterNames[i]) == null &&
					this.parameterDescriptorsAll.get( parameterNames[i]).defaultValue  == null ) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
											"ALDOperator.validateGeneric(): Required parameter \"" + parameterNames[i] + "\""+
                                    " is null in operator " + this.name);
			}
		} 
	}

	/**  Operator specific validation of parameters and inputs.
 	 */
	public void validateCustom() throws ALDOperatorException {
		return;
	}

	// ===================================================================
	/** Set the verbose state of this opertor.
	 * @param	verbose 	New verbose state
	 */
	public void setVerbose( Boolean verbose) throws ALDOperatorException {
		this.setSupplemental( "verbose", verbose);
	}

	/** Return the verbose state of this opertor.
	 * @return  Current verbose state
	 */
	public Boolean getVerbose() throws ALDOperatorException {
		return (Boolean)(this.getSupplemental( "verbose"));
	}

	// ===================================================================

	/** This method does the actual work and needs to be implemented by every subclass.
	 */

	protected abstract void operate() throws ALDOperatorException,ALDProcessingDAGException;

	// ===================================================================

	/** Write the parameter values to an xml file
	 *
	 * @param	filename	filename of the xml file to write to
	 */

	public void writeParametersToXml( String filename ) {
		try {
			PrintStream out = new PrintStream( filename);

			XStream xstream = new XStream(new DomDriver());
			//String xml = xstream.toXML( parameterHash);
			String xml = xstream.toXML( new ALDParameterWrapper(this));
	
			out.println( xml);
		} catch ( Exception e) {
			System.err.println(e);
			e.printStackTrace();
		}

	}

	/** parse the parameter values to an XmlObject
	 *
	 */

	public XmlObject parametersToXmlObject() 
		throws XmlException {

		XStream xstream = new XStream(new DomDriver());
		String xml = xstream.toXML( new ALDParameterWrapper(this));
		XmlObject res = XmlObject.Factory.parse( xml);
	
		return res;
	}

	/** Set the parameter values as read from an Xml file.
	 * The class and package name found in the Xml are compared to
	 * the operator to set the parameters to and are required to match.
	 *
	 * @param	filename	filename of the xml file to read from
	 * @return	version string found in Xml file of null if Xml file could not be properly read
	 *				or class or package name check failed
	 */

	public String setParametersFromXml( String filename ) {
		return setParametersFromXml( filename, getClass().getSimpleName(), getClass().getPackage().getName());
	}
		
	/** Set the parameter values as read from an Xml file.
	 * The class and package name found in the Xml are compared to
	 * the <code>className</code> and <code>packageName</code> respectively.
	 * If one of these names is <code>null</code> any name in the Xml file
	 * is accepted.
	 *
	 * @param	filename	filename of the xml file to read from
	 * @param	className	class name expected in the Xml file or null
	 * @param	packageName	package name expected in the Xml file or null
	 * @return	version string found in Xml file of null if Xml file could not be properly read
	 *				or class or package name check failed
	 */

	public String setParametersFromXml( String filename, String className, String packageName ) {
		
		try {
			FileInputStream in = new FileInputStream( filename);

			XStream xstream = new XStream(new DomDriver());

			ALDParameterWrapper pw = (ALDParameterWrapper)(xstream.fromXML( in));

			if ( className != null && ! className.equals( pw.getClassName()) ) {
				System.err.println( "ALDOperator::setParametersFromXml class name does not match, got " + 
					pw.getClassName() + " but expected " + className);
				return null;
			}

			if ( packageName != null && ! packageName.equals( pw.getPackageName()) ) {
				System.err.println( "ALDOperator::setParametersFromXml package name does not match, got " + 
					pw.getPackageName() + " but expected " + packageName);
				return null;
			}

			// we have to set the values directly
			String[] parameterNames = this.getParameterNames();
            for ( int i = 0; i < this.getNumParameters() ; i++ ) {
				try {
					setParameter( parameterNames[i], pw.getParameteres().get( parameterNames[i]));
				} catch ( ALDOperatorException e ) {
                       System.err.println( "ALDOperator::setParametersFromXml Error: cannot get value for parameter " +
                                       parameterNames[i]);
                       e.printStackTrace();
                }
			}

			return pw.getVersion();
			
		} catch ( FileNotFoundException e) {
			System.err.println( "ALDOperator::setParametersFromXml cannot open " + filename);
			return null;
		}

	}

	// ===================================================================

	/** Print some information of the current state this operator to System.out.
	 */

	public void print() {
		print( System.out);
	}
		
	/** Print some information the current state of this operator to <code>outfile</code>.
	 * @param	outfile	Stream to print on
	 */

	public void print( PrintStream outfile) {
		try {
			outfile.println( "  ALDOperator: " + this.name + " -> " + this); 

			String[] parameterNames = getParameterNames();
			for ( int n=0 ; n < getNumParameters() ; n++ )
				outfile.println( "    Parameter " + this.parameterDescriptorsAll.get( parameterNames[n]).name + " = " + getParameter( parameterNames[n]));

			String[] inputNames = getInputNames();
			for ( int n=0 ; n < getNumInputs() ; n++ )
				outfile.println( "    Input " + this.inputDescriptorsAll.get( inputNames[n]).name + " = " + getInput( inputNames[n]));

			String[] outputNames = getOutputNames();
			for ( int n=0 ; n < getNumOutputs() ; n++ )
				outfile.println( "    Output " + this.outputDescriptorsAll.get( outputNames[n]).name + " = " + getOutput( outputNames[n]));

			String[] supplementalNames = getSupplementalNames();
			for ( int n=0 ; n < getNumSupplementals() ; n++ )
				outfile.println( "    Supplemental " + this.supplementalDescriptorsAll.get( supplementalNames[n]).name + " = " + getSupplemental( supplementalNames[n]));

		} catch (ALDOperatorException e) {
			e.printStackTrace();
		}
	}

	/** Print information of the interface of this operator to System.out.
	 */

	public void printInterface() {
		printInterface( System.out);
	}
		
	/** Print information of the interface of this operator to <code>outfile</code>.
	 * @param	outfile	Stream to print on
	 */

	public void printInterface( PrintStream outfile) {
		outfile.println( "Interface of ALDOperator: " + this.name);

		String[] parameterNames = getParameterNames();
		for ( int n=0 ; n < getNumParameters() ; n++ )
			outfile.println( "    Param <" + this.parameterDescriptorsAll.get( parameterNames[n]).name +  "> " +
							(this.parameterDescriptorsAll.get( parameterNames[n]).required ? "" : " not") +
							" required, type: " + this.parameterDescriptorsAll.get( parameterNames[n]).myclass.getSimpleName());

		String[] inputNames = getInputNames();
		for ( int n=0 ; n < getNumInputs() ; n++ )
			outfile.println( "    Input <" + this.inputDescriptorsAll.get( inputNames[n]).name + "> " +
							(this.inputDescriptorsAll.get( inputNames[n]).required ? "" : " not") +
							" required, type: " + this.inputDescriptorsAll.get( inputNames[n]).myclass.getSimpleName());

		String[] outputNames = getOutputNames();
		for ( int n=0 ; n < getNumOutputs() ; n++ )
			outfile.println( "    Output <" + this.outputDescriptorsAll.get( outputNames[n]).name + "> " +
							(this.outputDescriptorsAll.get( outputNames[n]).required ? "" : " not") +
							" required, type: " + this.outputDescriptorsAll.get( outputNames[n]).myclass.getSimpleName());

		String[] supplementalNames = getSupplementalNames();
		for ( int n=0 ; n < getNumSupplementals() ; n++ )
			outfile.println( "    Supplemental <" + this.supplementalDescriptorsAll.get( supplementalNames[n]).name + "> " +
							(this.supplementalDescriptorsAll.get( supplementalNames[n]).required ? "" : " not") +
							" required, type: " + this.supplementalDescriptorsAll.get( supplementalNames[n]).myclass.getSimpleName());
	}

	// ================== helper function

	/** Is this class allowed as Input and/or Output?
	 */
	boolean allowedClassForIO( Class currentClass) {
		if ( this.debug ) {
			System.out.println( "       ALDOperator::allowedClassForIO for class " +
				currentClass.getName());
		}

		return ! ( currentClass == int.class || 
				   currentClass == byte.class || 
				   currentClass == short.class || 
				   currentClass == long.class || 
				   currentClass == float.class || 
				   currentClass == double.class || 
				   currentClass == boolean.class || 
				   currentClass == char.class) ;
	}

	/** Does any hashtable contain the key, i.e. this field?
	 */
	private boolean fieldContained( String key) {
		return this.parameterDescriptorsAll.containsKey( key) |
				 this.inputDescriptorsAll.containsKey( key) |
				 this.outputDescriptorsAll.containsKey( key) |
				 this.supplementalDescriptorsAll.containsKey( key) ;
	}
	

}

