/*
 * This file is part of Alida, a Java library for 
 * Automatic Logging of Process Information in Data Analysis.
 *
 * Copyright (C) 2010 - 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 3762 $
 * $Date: 2011-05-04 18:22:37 +0200 (Mi, 04 Mai 2011) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.tools;

/**
 * This class serves as helper class for reading and writing environment
 * variables and Java environment properties, respectively.
 * <p>
 * Every Alida operator and also every other class can define properties. 
 * To ensure a certain structure of the properties and avoid chaos in property 
 * names the properties should meet to the following convention:
 * 
 *  alida.operatorname.property
 * 
 * The corresponding environment variable will then be 
 * 
 * ALIDA_OPERATORNAME_PROPERTY
 * 
 * following common Unix/Linux conventions. 
 * <p>
 * 
 * @author moeller
 *
 */
public class ALDEnvironmentConfig {

	/**
	 * This function reads out the value of a given environment property.
	 * <p>
	 * Environment properties can be defined in terms of environment variables,
	 * or java properties passed to the virtual machine. The options are 
	 * checked in exactly this sorting. If the requested property is not found
	 * in either of the two configuration environments, null is returned.
	 * <p>
	 * The actual name of the property checked is composed of the given prefix,
	 * operator name and then environment variable name (in this order).
	 * For checking environment variables all strings are converted to 
	 * upper-case and linked together by inserting '_' in between.
	 * For checking JVM properties, dots are inserted in between and all strings
	 * are converted to lower-case.
	 * 
	 * @param _prefix				Prefix.
	 * @param _operator			Name of the operator.
	 * @param _envVariable	Name of the property.
	 * @return	Value of property, NULL if not existing.
	 */
	public static String getConfigValue(String _prefix, String _operator, 
			String _envVariable) {

		String prefix = (_prefix == null || _prefix.isEmpty()) ? 
																											"" : _prefix + "_";
		String operator = (_operator == null || _operator.isEmpty()) ? 
																											"" : _operator + "_";
		String envVariable = (_envVariable == null || _envVariable.isEmpty()) ? 
																											"" : _envVariable;

		String envVarNameTmp= prefix + operator + envVariable;
		String envVarName= envVarNameTmp.toUpperCase();
		
		prefix = (_prefix == null || _prefix.isEmpty()) ? 
																								"" : _prefix + ".";
		operator = (_operator == null || _operator.isEmpty()) ? 
																								"" : _operator + ".";
		envVariable = (_envVariable == null || _envVariable.isEmpty()) ? 
																								"" : _envVariable;

		String envPropNameTmp= prefix + operator + envVariable;
		String envPropName= envPropNameTmp.toLowerCase();
		
		// look for an environment variable
    String tmpValue= null, value= null;
		if ((tmpValue = System.getenv(envVarName)) != null) 
      value= tmpValue;
		// check java property
		else if ((tmpValue = System.getProperty(envPropName)) != null) 
      value= tmpValue;
		return value;
	}
	
	/**
	 * This function reads out the value of a given environment property.
	 * <p>
	 * Here the default prefix "alida" is assumed. It is combined with operator 
	 * and environment variable name in this order.
	 * 
	 * @param _operator			Name of the operator.
	 * @param _envVariable	Name of the property.
	 * @return	Value of property, NULL if not existing.
	 */
	public static String getConfigValue(String _operator, String _envVariable) {
		return getConfigValue("alida", _operator, _envVariable);
	}

	/**
	 * This function reads out the value of a given environment variable.
	 * <p>
	 * Default prefix is "alida".
	 * 
	 * @param _operator			Name of operator, ignored if null.
	 * @param _envVariable	Name of the variable.
	 * @return	Value of property, NULL if not existing.
	 */
	public static String getEnvVarValue(String _operator, String _envVariable) {
		String operator = (_operator == null || _operator.isEmpty()) ? 
																											"" : _operator + "_";
		String envVariable = (_envVariable == null || _envVariable.isEmpty()) ? 
																											"" : _envVariable;
		String envVarNameTmp= "alida_" + operator + envVariable;
		String envVarName= envVarNameTmp.toUpperCase();		
		return System.getenv(envVarName); 
	}
	
	/**
	 * This function reads out the value of a given JVM property.
	 * 
	 * @param _prefix				Prefix, ignored if null.
	 * @param _operator			Name of operator, ignored if null.
	 * @param _envVariable	Name of the property.
	 * @return	Value of property, NULL if not existing.
	 */
	public static String getJVMPropValue(String _prefix, String _operator, 
																				String _envVariable) {
		String prefix = (_prefix == null || _prefix.isEmpty()) ? 
																												"" : _prefix + ".";
		String operator = (_operator == null || _operator.isEmpty()) ? 
																												"" : _operator + ".";
		String envVariable = (_envVariable == null || _envVariable.isEmpty()) ? 
																												"" : _envVariable;

		String envPropNameTmp= prefix + operator + envVariable;
		String envPropName= envPropNameTmp.toLowerCase();
		return System.getProperty(envPropName);
	}
	
	/**
	 * This function reads out the value of a given JVM property.
	 * <p>
	 * Default prefix is "alida".
	 * 
	 * @param _operator			Name of operator, ignored if null.
	 * @param _envVariable	Name of the property.
	 * @return	Value of property, NULL if not existing.
	 */
	public static String getJVMPropValue(String _operator, String _envVariable){
		return getJVMPropValue("alida", _operator, _envVariable);
	}

	/**
	 * This function reads out the value of a given ImageJ preference.
	 * 
	 * @param plugin	Name of plugin, ignored if null.
	 * @param envVariable	Name of the preference.
	 * @return	value of property, null if not existing
	 */
//	public static String getImageJPropValue(String plugin, String envVariable) {
//		String prefix= null;
//		if (plugin == null)
//			prefix= "mitobo";
//		else
//			prefix= "mitobo." + plugin;
//		String envPropNameTmp= prefix + "." + envVariable;
//		String envPropName= envPropNameTmp.toLowerCase();
//		return ij.Prefs.get(envPropName,null);		
//	}
	
	/**
	 * This functions sets a given property to the specified value.
	 * <p>
	 * The properties are saved to ~/.imagej/Prefs.txt when the ImageJ gui is
	 * properly closed (not killed!) by the user. Note that already defined 
	 * property are simply overwritten with new values if the method is called 
	 * on existing properties.  
	 * 
	 * @param plugin	Name of Mitobo plugin, ignored if null.
	 * @param envVar	Name of property.
	 * @param val		Value of property.
	 */
//	public static void setImageJPref(String plugin, String envVar, String val) {
//		String prefix= null;
//		if (plugin == null)
//			prefix= "mitobo";
//		else
//			prefix= "mitobo." + plugin;
//
//		String envPropNameTmp= prefix + "." + envVar;
//		String envPropName= envPropNameTmp.toLowerCase();
//		ij.Prefs.set(envPropName, val);
//	}
}
