/*
 * This file is part of Alida, a Java library for 
 * Automatic Logging of Process Information in Data Analysis.
 *
 * Copyright (C) 2010 - 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 3762 $
 * $Date: 2011-05-04 18:22:37 +0200 (Mi, 04 Mai 2011) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.version;

import java.util.Hashtable;

import de.unihalle.informatik.Alida.tools.ALDEnvironmentConfig;

/**
 * Factory class dynamically instantiating version provider instances 
 * at runtime.
 * <p>
 * Here JVM properties are evaluated to dynamically configure the way, how 
 * Alida handles software versions. Usually there are different possibilities 
 * from where to get software version information. The probably most popular 
 * way is to query software repositories like CVS, SVN or Git. However, other
 * options can be imagined as well.
 * <p>
 * Alida supports dynamic configuration of the version data handling. This 
 * factory instantiates a concrete version provider based on the 
 * environment property <i>alida_versionprovider_class</i>. This property 
 * should contain the name of a class extending base class 
 * {@link ALDVersionProvider}. The generated instance of this class will be 
 * used for all software version requests triggered during Alida operator
 * invocations.  
 * 
 * @author moeller
 */
public class ALDVersionProviderFactory {
	
	/**
	 * Hashmap with global version provider objects.
	 */
	private static Hashtable<String, ALDVersionProvider> providerInstances = 
		new Hashtable<String, ALDVersionProvider>();
	
	/**
	 * Returns a reference to the provider object specified by the properties.
	 */
	public static ALDVersionProvider getProviderInstance() {
		String providerClass = getClassName();
		return getProviderInstance(providerClass);
	}
	
	/**
	 * Returns a reference to the provider object of the specified class.
	 */
	public static ALDVersionProvider getProviderInstance(String providerClass) {
		if (providerClass == null)
			return null;
		if (   !providerClass.isEmpty() 
				&& !providerInstances.containsKey(providerClass)) {
			Object instance = instantiateObject(providerClass);
			if (instance != null) {
				providerInstances.put(providerClass, (ALDVersionProvider)instance);
				return (ALDVersionProvider)instance;
			}
			return null;
		}
		return providerInstances.get(providerClass);
	}

	/**
	 * Instantiates an object of the specified version provider class.
	 */
	private static Object instantiateObject(String providerClass) {
		Object providerObject = null;
    try {
        Class<?> myclass = Class.forName(providerClass);
        providerObject = myclass.newInstance();
        
        // check if object is of correct type
        if (!(providerObject instanceof ALDVersionProvider)) {
        	System.err.println("ALDVersionProviderFactory: " +
        		"Cannot instantiate version provider object, invalid class!");
        	return null;
        }
    } catch (ClassNotFoundException ex) {
    	System.err.println("ALDVersionProviderFactory: " + 
    			               "Could not find class " + providerClass);
    	return null;
    } catch (IllegalAccessException ex) {
    	System.err.println("ALDVersionProviderFactory: " + 
    			               "Could not access class " + providerClass);
    	return null;
    } catch (InstantiationException ex) {
    	System.err.println("ALDVersionProviderFactory: " + 
    			               "Could not instantiate class " + providerClass);
    	return null;
    }
    return providerObject;
	}

	/**
	 * Returns version provdier class name from environment configuration.
	 */
	public static String getClassName() {
		return ALDEnvironmentConfig.getConfigValue("alida","","versionprovider_class");
	}
	
	/**
	 * Returns true if the factory is properly configured.
	 */
	public static boolean isClassNameSpecified() {
		if (ALDEnvironmentConfig.getConfigValue("alida","", 
				 																		"versionprovider_class") != null
			&& !ALDEnvironmentConfig.getConfigValue("alida","", 	
															"versionprovider_class").isEmpty()) {
			return true;
		}
		return false;
	}

	/**
	 * Test routine to check functionality of factory and provider classes.
	 * 
	 * @param args	Ignored.
	 */
	public static void main(String[] args) {
		System.out.println("Trying to get an instance...");
    ALDVersionProvider vp = getProviderInstance();
    if (vp != null)
    	System.out.println("Version is " + vp.getVersion());
    else
    	System.err.println("Something went wrong, provider object is null!");
    System.out.println("Done my work!");
	}
}
